from .toolbox import center, color, clear, underline
from .toolbox import generate_otp, timetools, label
from .toolbox import no_data, iter_print, wrap_text
from .toolbox import warning, choose, list_options
from datetime import datetime
from . import storage
import time

def menu():
    header = center(" 《 USER PROFILE 》 ", line="—")
    print(f"\n{header}\n")
    
    options = {
        "View profile": show_profile,
        "Edit profile fields": edit_profile,
        "Edit commandments": edit_commandments,
        "Reset profile": reset_profile
    }

    list_options(options, "Choose action")
    choose(options, src=menu)

def show_profile():
    data = storage.load_data(storage.USER_FILE)
    if not data:
        no_data("No user data found")
        underline()
        return

    labels = label(["Name:", "Birthday:", "Chief Aim:"])
    
    print(f"\n{labels[0]} {data.get('Name', 'N/A')}")
    bday = timetools.timestamp(data['Birthday'], 1)
    print(labels[1], bday)
    print(f"{labels[2]}\n")
    print(f"{wrap_text(data.get('Aim')) or 'N/A'}\n")

    if "Decalogue" in data:
        list_options(data["Decalogue"], "Commandments", 
                     spaced=True)

    underline()

def edit_profile():
    data = storage.load_data(storage.USER_FILE)
    
    fields = ["Name", "Birthday", "Timezone", "Aim"]
    field_names = {
        "Name": "Name",
        "Birthday": "Birthday (YYYY-MM-DD)",
        "Timezone": "Timezone (e.g. CAT)",
        "Aim": "Chief Aim"
    }

    print()
    while True:
        list_options(fields, "Which field do you want"
                            +" to edit?")
        print(f"{len(fields)+1}. Back")
        try:
            choice = int(input(color(">>> ", 
                     "magenta")))
            if choice == len(fields)+1:
                print()
                underline()
                break
            field = fields[choice - 1]
            
            print()
            while True:
                new_val = input(color(f"New value for"
                    f"{field_names[field].lower()}: ", 
                    "cyan")).strip()
                if choice == 2:
                    new_val = timetools.to_iso(
                              new_val)
                    if new_val == None:
                        warning("Invalid format! Try "
                               +"again.")
                        continue
                break
            
            data[field] = new_val
            storage.save_data(storage.USER_FILE, data)
            foot =center(" Updated Successfully ","—")
            print(f"\n{foot}\n")
        except (ValueError, IndexError):
            warning("Invalid choice! Try again.")

def edit_commandments():
    data = storage.load_data(storage.USER_FILE)
    
    options = [
        "New commandment list", 
        "Add to existing"
    ]
    
    print()
    list_options(options)
    choice = choose(options, 1, edit_commandments)
    if choice is None: return
    
    print("\n", color("Enter new commandment(s). Type "
        +"'done' when finished:", "yellow"), "\n")
    
    new_decalogue = []
    if choice == 1: new_decalogue = data["Decalogue"]
        
    while True:
        cmd = input(color("Commandment "
            + f"{len(new_decalogue)+1}: ", 
              "cyan")).strip()
        if cmd.lower() == "done": break
        new_decalogue.append(cmd)
    
    data["Decalogue"] = new_decalogue
    storage.save_data(storage.USER_FILE, data)
    footer = center(" Commandments Updated ", "—")
    print(f"\n{footer}")

def reset_profile():
    otp    = generate_otp()
    prompt = "This will erase your profile and wipe " \
           + f"out your coffers. Type {otp!r} to confirm"
    print("\n" + color(wrap_text(prompt), "red"))
    confirm = input(color('>>> ', 'red'))
    if confirm == otp:
        storage.delete(storage.USER_FILE)
        storage.delete(storage.COFFER)
        time.sleep(1)
        print("\nReseting", end="")
        time.sleep(.5)
        iter_print(".", times=3, end="", delay=.75)
        time.sleep(.75)
        print()
        time.sleep(.5)
        footer = center(" Profile Reset ", "—")
        print(f"\n{footer}")        
    else: print("\n"+center(" Cancelled ", "—"))



"""
Module for managing "missions" (goal/task progress 
tracking)
"""
from .toolbox import list_options, past, color, center
from .toolbox import warning, clear, underline, choose
from .toolbox import no_data, rated_bar, spacer
from .toolbox import wrap_text as wrap
from . import slight_edge
from . import storage

def menu():
    header = center("《 Mission Center 》", "—")
    print(f"\n{header}\n")
    
    options = {
        "Update mission progress": mission_log,
        "Create new mission": create_mission
    }
    list_options(options, "Choose action")
    choose(options, src=menu)

def mission_log():
    """
Allows user to update progress on a mission (task).
If all objectives in a mission are complete, it marks it as 'Complete'.
    """
    data = storage.load_data(storage.TASK_FILE)
    if not data:
        no_data("No missions found")
        underline()
        return
    
    header = center(" Updating Task! ", "—")
    print(f"\n{header}\n")
    
    missions = [m["Mission"] for m in data]        
    
    list_options(missions, "Update mission")    
    choice = choose(missions, 1, mission_log)
    if choice is None: return
        
    current = data[choice]["Current"]      
    try:
        objective = data[choice]["Objectives"][
                    current]
        objective = past(objective)
    except IndexError:
        for mission in data:
           if mission["Mission"] == missions[choice]:
               mission["Status"]  = "Complete"
               storage.save_task(data)
               return

    probe = input(f"\nHave you {objective}? [y/n]: "
            ).lower().strip()
    if probe == "clear":
        clear()
        mission_log()
    elif probe == "y":
        slight_edge.ballot(1)
        for mission in data:
            if mission["Mission"] == missions[choice]:
                mission["Current"] += 1
                storage.save_data(storage.TASK_FILE, 
                    data)                
    print()
    underline()
    
def create_mission():
    print()
    print(center(" Create New Mission ", "—"))

    while True:
        name = input(color("\nMission name: ", 
               "cyan")).strip()
        if name: break
        warning("Mission name cannot be empty.")

    objectives = []
    print(color("\nEnter mission objectives one by "
         +"one. Type 'done' when finished:\n", 
         "yellow"))
    
    while True:
        objective = input(color("Objective "
                  + f"{len(objectives)+1}: ", 
                    "green")).strip()
        if objective.lower() == "done":
            if not objectives:
                warning("At least one objective is "
                       +"required.")
                continue
            break
        elif not objective:
            warning("Objective cannot be empty.")
            continue
        objectives.append(objective)

    mission = {
        "Mission": name,
        "Objectives": objectives,
        "Current": 0,
        "Status": "Incomplete"
    }

    storage.save_entry(storage.TASK_FILE, mission)
    print("\n",center(" ✔ Mission Created! ", "—"))
    
def get_missions() -> list[dict]:
    """
Returns a list of mission summaries (progress, name, current objective).
    """
    data     = storage.load_data(storage.TASK_FILE)
    missions = []
    for mission in data:
        if mission["Status"] != "Complete":
            current = mission["Current"]
            rate = current/len(mission["Objectives"])
            obj  = mission["Objectives"][current]
            info = {
                "Name": mission["Mission"],
                "Progress": round(rate, 4),
                "Objective": obj
            }
            missions.append(info)
            
    return missions[:min(len(missions), 3)]

def render_mission_card(mission: dict, index: int):
    name      = mission["Name"]
    objective = mission["Objective"]
    rate      = mission["Progress"]
    bar       = rated_bar(rate)
    rated     = center(f'{bar}')
    
    spacer(2)
    m_tag = color(f'Mission {index+1}:', 'lightcyan')
    o_tag = color(f'Objective:', 'cyan')
    print(wrap(f"{m_tag} {name}", indent=11))
    print(wrap(f"{o_tag} {objective}\n", indent=11))
    RATE = center(" REALIZATION RATE ")
    print(f"\n{RATE}")
    print(f"{rated}")
    spacer(2)


from .toolbox import center, choose, list_options
from . import stat_logic
from . import about

def menu():
    header = center("《 Settings 》", line="—")
    print(f"\n{header}\n")
    
    options = {
        "Stats": stat_logic.menu,
        "About": about.menu
    }
    
    list_options(options)
    choose(options, src=menu)


from .toolbox import clear, color, center, underline
from .toolbox import warning, timetools, iter_print
from .toolbox import wrap_text, help_msg
from . import storage
import time
import sys
import os

USER_FILE_PATH = storage.DATA_DIR / storage.USER_FILE

def user_exists():
    if os.path.exists(USER_FILE_PATH):
        with open(USER_FILE_PATH) as f:
            if len(list(f)) > 1: return True

def commandments():
    msg = "Enter commandments one by one." \
        + "Type 'done' when finished:"
    print("\n"+color(wrap_text(msg), 'yellow'), "\n")
    
    decalogue = []
    while True:
        commandment = input(color(
            f"Commandment {len(decalogue)+1}: ",
            "cyan")).strip()
        
        # Stops input if user types 'done'
        if commandment.lower() == "done": break
        elif commandment: decalogue.append(commandment)
    
    return decalogue

def setup_user():
    print("\n"+center(" 《 HABITRAX SETUP 》 ", "="))
    welcome = f"Welcome! Let's get a few details to " \
            + f"personalize your experience."    
    print(f"\n{wrap_text(welcome)}\n")

    user_data = {}
    user_data["Name"] = input(color(
        "Your name or nickname: ", "cyan")).strip()
    
    name = user_data.get("Name", "Anonymous")
    while True:
        try: 
            birthday = input(color("Your birthday (YYYY"
                     + "-MM-DD): ", "cyan")).strip()
            user_data["Birthday"] = timetools.to_iso(
                                    birthday)
            if user_data["Birthday"] is None:
                raise ValueError
            break
        except ValueError:
            warning("Invalid format! Try again.")
    
    user_data["Aim"] = input(color("Your chief aim: ", 
                       "cyan")).strip()    
    prompt = input("\nDo you want to write your "
           + "commandments? [y/n]\n"
           + f"{color('>>> ', 'magenta')}")
    
    if prompt and prompt[0].lower() == "y":
        user_data["Decalogue"] = commandments()
    
    if name:
        print("\n"+color(f"Welcome, {name}!", "green"))
    storage.save_data(storage.USER_FILE, user_data)
    footer = center(" Setup complete! ", "—")
    print(f"\n{footer}\n")
    time.sleep(.75)
    print(color("Launching Habitrax", "green"), end="", 
        flush=True)
    time.sleep(1)
    iter_print(color(".", "green"), times=3, end="", 
        delay=.75)
    time.sleep(1)
    clear(print_header=False)

if __name__ != "__main__": 
    if not user_exists():
        if len(sys.argv) > 1: help_msg(True, True)
        try: setup_user()
        except (KeyboardInterrupt, EOFError) as e:
            if isinstance(e, EOFError): print()
            warning("Setup cancelled by user.", True)
            underline()
            time.sleep(1)
            clear(print_header=False)
            sys.exit(1)



from .toolbox import choose_from, var_per_cat, choose
from .toolbox import warning, list_options, wrap_text
from .toolbox import timetools, label, underline, pc
from .toolbox import color, center, no_data, clear 
from .toolbox import make_progress_bar as mpb
from tuikit.textools import strip_ansi
from datetime import datetime as dt
from operator import itemgetter
from . import storage

catch = []

def menu(from_term: bool = False):
    header = center("《 Statistics 》", "—")
    print(f"\n{header}")
    
    data = storage.load_data(storage.DATA_FILE)
    if not data:
        no_data("No data found")
        underline()
        return
        
    options = {
        "Overall stats": overall_stats,
        "Stats by category": by_category,
        "Stats by time of day": by_time,
        "All records": view_all,
        "View levels": get_levels
    }
    
    print()
    list_options(options, "View")    
    choose(options, src=menu)
    if from_term: print()

def overall_stats():
    header = center(" Overall ")
    print(f"\n{color(header, fg='green')}\n")
    
    data       = storage.load_data(storage.DATA_FILE)
    total      = len(data)
    total_time = sum(entry['Time spent'] for entry 
                 in data)
    avg_focus  = round(sum(entry['Focus'] for entry 
                 in data) / total, 2)
    avg_q_min  = round(sum(entry['Quality range'][0] 
                 for entry in data) / total, 2)
    avg_q_max  = round(sum(entry['Quality range'][1] 
                 for entry in data) / total, 2)
    avg_duration = timetools.format_time(total_time, 
                   total)
    total_time   = timetools.format_time(total_time)       
    time_pc, focus_pc, range_pc = pc.get_percent()   
    
    labels = label([
        "Avg focus:", "Total time:", 
        "Avg time:", "Total sessions:", 
        "Avg quality range:"
    ])    
    
    avg_fcs = f"{labels[0]} {avg_focus} {focus_pc}"
    avg_drn = f"{labels[2]} {avg_duration} {time_pc}"
    fcs     = len(strip_ansi(labels[0])) + 1
    drn     = len(strip_ansi(labels[2])) + 1
    
    print(wrap_text(avg_fcs, indent=fcs))
    print(wrap_text(avg_drn, indent=drn))
    print(labels[1], total_time)
    print(labels[3], total)       
    print(labels[4], avg_q_min,"—",avg_q_max,range_pc)    
    print()
    underline()

def by_category():
    header = center(" By Category ")
    print(f"\n{color(header, fg='green')}\n")
    
    file = storage.DATA_FILE
    categories = storage.get_tables(file, repair=False)
    cats_reprd = storage.get_tables(file)  
    
    labels = label([
        "Category:", "Avg Time:", "Avg Focus:",
        "Total time:", "Total sessions:"
    ])
    
    for cat_reprd, cat in zip(cats_reprd, categories):
        entries    = storage.load_table_as_list(file, cat)
        total      = len(entries)
        avg_focus  = sum(e["Focus"] for e in entries)\
                   / total
        total_time = sum(e["Time spent"] for e in entries)
        avg_duration      = timetools.format_time(
                            total_time, total)
        total_duration    = timetools.format_time(
                            total_time)
        time_pc, focus_pc = pc.get_grouped_percent(entries)
                
        avg_drn = f"{labels[1]} {avg_duration} "\
                + f"{time_pc}"
        avg_fcs = f"{labels[2]} {avg_focus:.2f} "\
                + f"{focus_pc}"
        drn     = len(strip_ansi(labels[1])) + 1
        fcs     = len(strip_ansi(labels[2])) + 1
    
        print(labels[0], cat_reprd)
        print(wrap_text(avg_drn, indent=drn))
        print(wrap_text(avg_fcs, indent=fcs))
        print(labels[3], total_duration)
        print(labels[4], total, end="\n\n")               
    underline()

def by_time():
    header = center(" By Time of Day ")
    print(f"\n{color(header, fg='green')}\n")
    
    data = storage.load_data(storage.DATA_FILE)
    
    timeblock = {
        'Morning': [], 'Afternoon': [], 
        'Evening': [], 'Night': []
    }

    for entry in data:
        hr = dt.fromisoformat(entry["Timestamp"]).hour
        if 5 <= hr < 12:
            timeblock['Morning'].append(entry)
        elif 12 <= hr < 17:
            timeblock['Afternoon'].append(entry)
        elif 17 <= hr < 21:
            timeblock['Evening'].append(entry)
        else: timeblock['Night'].append(entry)          
    
    labels = label([
        "Period:", "Avg Time:", "Avg Focus:", 
        "Total time:", "Total sessions:"
    ])
        
    for period, entries in timeblock.items():
        if not entries: continue
        total      = len(entries)
        avg_focus  = sum(e["Focus"] for e in entries
                 ) / total
        total_time = sum(e["Time spent"] for e in 
                     entries)
        total_duration    = timetools.format_time(
                            total_time)
        avg_duration      = timetools.format_time(
                            total_time, total)
        time_pc, focus_pc = pc.get_grouped_percent(
                            entries)
        
        avg_drn = f"{labels[1]} {avg_duration} "\
                + f"{time_pc}"
        avg_fcs = f"{labels[2]} {avg_focus:.2f} "\
                + f"{focus_pc}"
        drn     = len(strip_ansi(labels[1])) + 1
        fcs     = len(strip_ansi(labels[2])) + 1
    
        print(labels[0], period)
        print(wrap_text(avg_drn, indent=drn))
        print(wrap_text(avg_fcs, indent=fcs))
        print(labels[3], total_duration)
        print(labels[4], total, end="\n\n")
    underline()   

def category_records() -> list[dict] | None:
    file       = storage.DATA_FILE
    categories = storage.get_tables(file)
    print()
    list_options(categories, "Choose category")
    choice = choose(categories, 1, category_records)
    catch.append(choice)
    choice = catch[0]
    if choice is not None:        
        category = categories[choice]
        return storage.load_table_as_list(file, category)

def period_records() -> list[dict] | None:
    data = storage.load_data(storage.DATA_FILE)
    if not data:
        no_data("No data found")
        return
    
    timeblock = {'Morning': [], 'Afternoon': [], 
                 'Evening': [],     'Night': []}

    for entry in data:
        hr = dt.fromisoformat(entry["Timestamp"]).hour
        if 5 <= hr < 12:
            timeblock['Morning'].append(entry)
        elif 12 <= hr < 17:
            timeblock['Afternoon'].append(entry)
        elif 17 <= hr < 21:
            timeblock['Evening'].append(entry)
        else: timeblock['Night'].append(entry)

    print()    
    list_options(timeblock, "Choose period")    
    choice = choose(timeblock, 1, period_records)
    catch.append(choice)
    choice = catch[0]
    if choice is not None:
        return choose_from(timeblock, choice)

def view_all():
    header = center(" Viewing All Records! ")
    print(f"\n{color(header, fg='green')}")
    print()
    
    data = storage.load_data(storage.DATA_FILE)
    
    records = {
        "View All": "Placeholder", 
        "View Category": category_records, 
        "View Period": period_records
    }    
    list_options(records)
    choice = choose(records, 1, view_all)
    if choice is None: return
    elif choice > 0: 
        data = choose_from(records, choice)()
    if not data: return
    
    filters = {
        "A-Z": sorted(data, key=itemgetter('Task')),
        "Z-A": sorted(data, key=itemgetter('Task'), 
               reverse=True),
        "Latest": data.copy()[::-1],
        "Oldest": data,
        "Highest Focus": sorted(data, key=itemgetter(
                         'Focus'), reverse=True),
          "Least Focus": sorted(data, key=itemgetter(
                         'Focus')),
         "Longest Duration": sorted(data, key=itemgetter(
                             'Time spent'), reverse=True),
        "Shortest Duration": sorted(data, key=itemgetter(
                             'Time spent'))
    }
    
    print()
    list_options(filters, "Filter by")            
    choice = choose(filters, 1, view_all)
    if choice is None: return
    
    filtered = choose_from(filters, choice)            
    labels   = label([
        "Date:", "Task:", "Focus:",
        "Duration:", "Logged:"
    ])
    for entry in filtered:
        duration = timetools.format_time(entry[
                   'Time spent'])
        time = timetools.timestamp(entry['Timestamp'])
        lapsed = timetools.format_time_passed(entry[
                 'Timestamp'])
        description = f"{labels[1]} {entry['Task']}"
        print()
        print(labels[0], time)
        print(wrap_text(description, 6))
        print(labels[2], entry['Focus'])
        print(labels[3], duration)
        print(labels[4], lapsed, end="\n\n")
    underline()

def get_levels():
    header = center(" Skill Levels ", "—")    
    print(f"\n{header}\n")
    
    data = storage.load_data(storage.LEVEL_FILE)
    
    labels = label([
        "Category:", "Level:", "EXP",
        "Leveled up:", "On:"
    ])

    vpc = var_per_cat
    for index, entry in enumerate(data):
        if vpc(index) is None: continue
        bar, date, lapsed, exp_pc = vpc(index)        
        print(labels[0], entry['Category'])
        print(labels[1], entry['Level'])
        print(labels[2], bar, f"{exp_pc:.2f}%")
        print(labels[3], lapsed)
        print(labels[4], date, end="\n\n")
    underline()


from . import logger, status_window, debrief
from . import stats, soul_work, toolbox
from .main import main as menu
import argparse
import sys

def parse_args() -> argparse.ArgumentParser:
    parser         = argparse.ArgumentParser(description=
                     toolbox.help_msg())
    subparsers     = parser.add_subparsers(dest="command")
    session        = subparsers.add_parser("log")
    view           = subparsers.add_parser("view")
    bank           = subparsers.add_parser("coffer")
    
    comparsers = {
       "adders": [session.add_argument,
                  view.add_argument,
                  bank.add_argument],
       
       "cargs": [# session parser
                 [["-t", "--task"],
                  ["-r", "--reflect"], 
                  ["-s", "--shadow"]], 
                
                # view parser
                 [["-d", "--debrief"], 
                  ["-w", "--status-window"],
                  ["-s", "--stats"],
                  ["-b", "--balance"], 
                  ["-t", "--transactions"], 
                  ["-a", "--account"]], 
                
                # bank parser
                 [["-d", "--deposit"],
                  ["-w", "--withdraw"]]
                ]
    }
    
    adders, cargs = comparsers["adders"],comparsers["cargs"]
    for add_arg, args in zip(adders, cargs):
        for arg in args:
            short, long = arg
            add_arg(short, long, action="store_true")
    
    return parser.parse_args()

def main():
    args = parse_args()
    if not args.command: sys.exit(menu())
    
    print(args)
    if args.command == "log":
        if args.reflect: debrief.reflect(True)
        elif  args.task: logger.log_session(True)
        else           : debrief.shadow_journal(True)
    elif args.command == "view":
        if   args.debrief: debrief.debrief_review(0, 1)
        elif args.status_window: status_window.window(1)
        elif   args.stats: stats.menu(True)
        elif args.balance: soul_work.coffer(view=0)
        elif args.account: soul_work.coffer(view=1)
        else             : soul_work.coffer(view=2)
    else:
        if args.withdraw: soul_work.coffer(cashflow=0)
        else            : soul_work.coffer(cashflow=1)

if __name__ == "__main__": main()



"""
Logs work sessions, calculates quality ranges, and updates user progress (leveling, trajectory)
"""
from datetime import datetime
from . import slight_edge
from . import toolbox
from . import storage
from . import level
    
def log_session(from_term: bool = False):
    """
Logs a single work session with required metadata:
         Task description: Describe what you were doing
                 Category: What were you doing?
                           Reading? Studying?
       Focus level (1–10): Focus level or workout 
                           intensity 
Fuzziness/Uncertainty (±): How much did focus/intensity 
                           flactuate or how unsure are 
                           you of the rating you gave 
               Time spent: How long the session took 
                           (in hours)
Then calculates the quality range and stores all this in a JSON entry.
    """
    header = toolbox.center("《 New Session Log 》","—")
    print(f"\n{header}\n")
    
    var    = []    
    labels = toolbox.label([
        "Task description: ", 
        "Category: ", 
        "Focus level (1–10): ", 
        "Uncertainty in focus (±): ", 
        "Time spent (in hours): "
    ])
    
    err = (ValueError, KeyboardInterrupt, EOFError)
    for pos, label in enumerate(labels):
        while True:
            try:
                variable = input(label)
                if pos < 2:
                    if variable == "clear":
                        toolbox.clear()
                        log_session()
                    break
                elif pos > 1 and float(variable) >= 0:
                    variable = float(variable)
                    break
            except err as e:
                if isinstance(e, toolbox.ExitError()):
                    if isinstance(e, EOFError): print()
                    print()
                    toolbox.underline()
                    return
                elif variable == "clear":
                    toolbox.clear()
                    log_session()
            toolbox.warning("Invalid input. Please "
                    +"enter numbers where required.")    
        var.append(variable)

    
    task, category, focus, fuzziness, time_spent = var

    focus     = max(0, min(focus,     10))
    fuzziness = max(0, min(fuzziness,  2))

    q_min   = max( 0, (focus - fuzziness) * time_spent)
    if q_min > 10: q_min = 10
    q_max   = min(10, (focus + fuzziness) * time_spent)
    if q_max <  0: q_max =  0
    q_range = [q_min, q_max]

    session = {
            "Timestamp": datetime.now().isoformat()   ,
                 "Task": task                         ,
             "Category": category                     ,
                "Focus": focus                        ,
            "Fuzziness": fuzziness                    ,
           "Time spent": time_spent                   ,
        "Quality range": [round(q, 2) for q in q_range]
    }

    storage.save_entry(storage.DATA_FILE, session)
    slight_edge.ballot(1)
    level.Stat(category).level_up()
    footer = toolbox.center(" ✔ Session Logged "
           + "Successfuly! ", "—")
    print(f"\n{footer}")
    if from_term: print()


from .toolbox import no_data, color, center, underline
from .toolbox import get_exp, number_padding, warning
from .toolbox import var_per_cat, label, choose
from .toolbox import list_options, get_exp
from . import storage

def menu():   
    header = center(" Stat logic ")
    print(f"\n{color(header, 'green')}\n")
    
    options = {
        "Rename stats": rename,
        "Combine stats": combine
    }
        
    list_options(options)
    choose(options, src=menu)       
    
def rename():
    data = storage.load_data(storage.LEVEL_FILE)[1:]
    if not data:
        no_data("No stats found")
        return
    
    mapping  = storage.load_map()
    combined = mapping.get("Combinations", {})
    
    print()
    names = []
    for entry in data:
        name = entry["Category"]
        for co_name, group in combined.items():
            if name in group: name = co_name
        if name not in names: names.append(name)
        
    list_options(names)
    choice = choose(names, 1, rename)        
    if choice is None: return
        
    name     = names[choice]
    new_name = input(color("\nEnter new name for "
             + f"{name.lower()}: "))
    rename_stat(name, new_name)
    ft = center(f" {name} Renamed Successfully ", "—")
    print(f"\n{ft}")

def combine():
    data = storage.load_data(storage.LEVEL_FILE)[1:]
    if not data:
        no_data("No stats found")
        return
    
    mapping  = storage.load_map()
    combined = mapping.get("Combinations", {})   
    labels   = label([
        "Enter new name: ",
        "Choose stats to combine (e.g., 1 2): "
    ])
    
    print()       
    names = []
    for entry in data:
        name = entry["Category"]
        for co_name, group in combined.items():
            if name in group: name = co_name
        if name not in names:
            names.append(name.strip().title())
        
    list_options(names, "Choose name(s)")
    print(f"{len(names)+1}. Back\n")        
    while True:
        try:
            prompt = input(labels[1])
            if prompt == str(len(names)+1):
                print()
                underline()
                return
                
            choices = prompt.split()              
            indices = [int(num) for num in choices]
            if any(0 >= i or i > len(names) for i in 
                indices): raise ValueError
            break
        except ValueError: pass
        warning("Invalid input! Try again.")
        
    source_list = [names[int(c)-1] for c in choices]
        
    print()
    names = ["New"]
    if mapping["Combinations"]:
        names = [name for name in 
                mapping["Combinations"]]
        names.append("New")
    list_options(names, "Add to combination")
    choice = choose(names, 1, combine)
    if choice is None: return
                
    target_name = names[choice].title().strip()
    if target_name == "New":
        print()
        target_name = input(labels[0]).strip().title()
        
    combine_stats(target_name, source_list)
    footer = center(f" {len(source_list)} Names "
           + "Combined Successfully ", "—")
    print(f"\n{footer}")

def rename_stat(original: str, new_name: str):
    mapping = storage.load_map()
    mapping["Renames"][original] = new_name
    storage.save_data(storage.STAT_MAP, mapping)

def combine_stats(target_name:str, source_list:list):
    mapping      = storage.load_map()
    combinations = mapping["Combinations"]
    
    if target_name in combinations:
        combinations[target_name].extend(source_list)
        storage.save_data(storage.STAT_MAP, mapping)
        return
    
    combinations[target_name] = source_list
    storage.save_data(storage.STAT_MAP, mapping)

def get_resolved_stats() -> dict:
    data     = storage.load_data(storage.DATA_FILE)
    mapping  = storage.load_map()   
    renames  = mapping.get("Renames", {})
    combined = mapping.get("Combinations", {})
    resolved = {}

    for stat in data:
        original = stat["Category"]
        name     = renames.get(original, original)

        # If this stat is part of a combination
        for combined_name, group in combined.items():
            if original in group:
                name = combined_name
                name = renames.get(name, name)               

        resolved.setdefault(name, [])
        if original not in resolved[name]:
            resolved[name].append(original)

    return resolved
